from typing import TYPE_CHECKING, Any, Type

from deker.ABC.base_factory import BaseAdaptersFactory
from deker.ctx import CTX
from deker.uri import Uri

from deker_server_adapters.array_adapter import ServerArrayAdapter
from deker_server_adapters.collection_adapter import ServerCollectionAdapter
from deker_server_adapters.consts import STATUS_OK
from deker_server_adapters.errors import DekerServerError
from deker_server_adapters.httpx_client import HttpxClient
from deker_server_adapters.varray_adapter import ServerVarrayAdapter


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseCollectionAdapter, BaseStorageAdapter, BaseVArrayAdapter


class AdaptersFactory(BaseAdaptersFactory):
    """Factory that produces server adapters."""

    uri_schemes = ("http", "https")

    def __init__(self, ctx: "CTX", uri: "Uri") -> None:
        # Make base url, so all urls will be relative
        kwargs = {
            "base_url": uri.raw_url,
            "verify": True,
            "http2": True,
            "timeout": None,
        }

        # If Uri contains auth params, remember them
        if uri.username:
            kwargs.update({"auth": (uri.username, uri.password)})

        # update values from context kwargs
        kwargs.update(ctx.extra.get("httpx_conf", {}))

        # Instantiate an httpx client
        self.httpx_client = HttpxClient(**kwargs)

        # We have to copy ctx to create new instance of extra
        # so client would be different for every factory
        copied_ctx = CTX(
            ctx.uri,
            config=ctx.config,
            storage_adapter=ctx.storage_adapter,
            executor=ctx.executor,
            is_closed=ctx.is_closed,
        )

        copied_ctx.extra["httpx_client"] = self.httpx_client
        self.do_healthcheck()
        super().__init__(copied_ctx, uri)

    def close(self) -> None:
        """Shutdown executor and httpx client."""
        self.httpx_client.close()
        super().close()

    def get_array_adapter(
        self,
        collection_path: Uri,
        storage_adapter: Type["BaseStorageAdapter"],
        *args: Any,
        **kwargs: Any,
    ) -> "BaseArrayAdapter":
        """Make server array adapter.

        :param storage_adapter: Type of storage adapter
        :param collection_path: URI of collection
        :param args: Won't be passed further
        :param kwargs: Won't be passed further
        """
        return ServerArrayAdapter(
            collection_path=collection_path,
            ctx=self.ctx,
            executor=self.executor,
            storage_adapter=storage_adapter,
        )

    def get_varray_adapter(
        self,
        collection_path: Uri,
        storage_adapter: Type["BaseStorageAdapter"],
        *args: Any,
        **kwargs: Any,
    ) -> "BaseVArrayAdapter":
        """Make server varray adapter.

        :param storage_adapter: Type of storage adapter
        :param collection_path: URI of collection
        :param args: Won't be passed further
        :param kwargs: Won't be passed further
        """
        return ServerVarrayAdapter(
            collection_path=collection_path,
            ctx=self.ctx,
            executor=self.executor,
            storage_adapter=storage_adapter,
        )

    def get_collection_adapter(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> "BaseCollectionAdapter":
        """Make server collection adapter.

        :param args: Won't be passed further
        :param kwargs: Won't be passed further
        """
        return ServerCollectionAdapter(self.ctx)

    def do_healthcheck(self) -> None:
        """Check if server is alive."""
        try:
            response = self.httpx_client.get("/v1/ping")
        except Exception:
            self.httpx_client.close()
            raise DekerServerError(
                None,
                "Healthcheck failed. Server is unavailable. Deker client will be closed.",
            )

        if response.status_code != STATUS_OK:
            self.httpx_client.close()
            raise DekerServerError(
                response,
                "Healthcheck failed. Deker client will be closed.",
            )



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING, Type, Union

from deker.ABC.base_adapters import BaseStorageAdapter, BaseVArrayAdapter
from deker.ctx import CTX

from deker_server_adapters.base import ServerArrayAdapterMixin
from deker_server_adapters.consts import ArrayType


if TYPE_CHECKING:
    from deker.uri import Uri


class ServerVarrayAdapter(ServerArrayAdapterMixin, BaseVArrayAdapter):
    """Server adapter for varrays."""

    type = ArrayType.varray

    def __init__(
        self,
        collection_path: Union["Uri", Path],
        ctx: CTX,
        executor: ThreadPoolExecutor,
        storage_adapter: Type["BaseStorageAdapter"],
    ) -> None:
        super().__init__(
            collection_path,
            ctx,
            executor,
            storage_adapter,
        )
        self.executor = executor



from collections.abc import Generator
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np

from deker import Array, Collection, VArray
from deker.ABC.base_array import BaseArray
from deker.tools.time import convert_datetime_attrs_to_iso
from deker.types import ArrayMeta, Numeric, Slice
from deker.uri import Uri
from deker_tools.slices import slice_converter
from httpx import Client, TimeoutException
from numpy import ndarray

from deker_server_adapters.consts import NOT_FOUND, STATUS_CREATED, STATUS_OK, TIMEOUT, ArrayType
from deker_server_adapters.errors import DekerServerError, DekerTimeoutServer


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.ABC.base_schemas import BaseArraysSchema


def create_array_from_meta(
    type: ArrayType,
    *args: Any,
    **kwargs: Any,
) -> Union[Array, VArray]:
    """Override class method.

    :param type: Type of Array
    :param args: Create array meta params
    :param kwargs: Create array meta kwargs
    """
    class_ = type.value
    return class_._create_from_meta(*args, **kwargs)


class ServerAdapterMixin:
    """Mixin for network communication."""

    @property
    def client(self) -> Client:
        """Return client singleton."""
        # We don't need to worry about passing args here, cause it's a singleton.
        return self.ctx.extra["httpx_client"]  # type: ignore[attr-defined]


class ServerArrayAdapterMixin(ServerAdapterMixin):
    """Mixin with server logic for adapters."""

    type: ArrayType
    collection_path: Uri

    def create(self, array: Union[dict, "BaseArray"]) -> BaseArray:
        """Create array on server.

        :param array: Array instance
        :return:
        """
        response = self.client.post(
            f"{self.collection_path.raw_url}/{self.type.name}s",
            json={
                "primary_attributes": convert_datetime_attrs_to_iso(
                    array["primary_attributes"],
                ),
                "custom_attributes": convert_datetime_attrs_to_iso(
                    array["custom_attributes"],
                ),
            },
        )
        if response.status_code != STATUS_CREATED:
            raise DekerServerError(response, "Couldn't create an array")

        try:
            data = response.json()
        except JSONDecodeError:
            raise DekerServerError(response, "Couldn't parse json")

        instance_id = data.get("id")
        if not instance_id:
            raise DekerServerError(response, "Server response doesn't contain ID field")

        kwargs = {
            "collection": array["collection"],
            "adapter": array["adapter"],
            "id_": instance_id,
            "primary_attributes": array["primary_attributes"],
            "custom_attributes": array["custom_attributes"],
        }

        model = self.type.value
        if model == VArray:
            kwargs["array_adapter"] = array["array_adapter"]
        return model(**kwargs)

    def read_meta(self, array: "BaseArray") -> ArrayMeta:
        """Read metadata of (v)array.

        :param array: Instance of (v)array
        :return:
        """
        response = self.client.get(
            f"{self.collection_path.raw_url}/{self.type.name}/by-id/{array.id}",
        )
        if response.status_code != STATUS_OK:
            raise DekerServerError(response, "Couldn't fetch an array")
        return response.json()

    def update_meta_custom_attributes(
        self,
        array: "BaseArray",
        attributes: dict,
    ) -> None:
        """Update custom attributes of (v)array.

        :param array: Instance of (v)array
        :param attributes: dict with attributes to update
        """
        response = self.client.put(
            f"{self.collection_path.raw_url}/{self.type.name}/by-id/{array.id}",
            json={"custom_attributes": convert_datetime_attrs_to_iso(attributes)},
        )
        if response.status_code != STATUS_OK:
            raise DekerServerError(response, "Couldn't update attributes")

    def delete(self, array: "BaseArray") -> None:
        """Delete array on server.

        :param array: Array/Varray to be deleted
        """
        resource = self.client.delete(f"{self.collection_path.raw_url}/{self.type.name}/by-id/{array.id}")
        if resource.status_code != STATUS_OK:
            raise DekerServerError(resource, f"Couldn't delete the {self.type.name}")

    def read_data(
        self,
        array: "BaseArray",
        bounds: Slice,
    ) -> Union[Numeric, ndarray, None]:
        """Read data from array/varray.

        :param array: From what array we read
        :param bounds: What part of array we read
        :return:
        """
        bounds_ = slice_converter[bounds]
        try:
            response = self.client.get(
                f"/v1/collection/{array.collection}/{self.type.name}/by-id/{array.id}/subset/{bounds_}/data",
                headers={"Accept": "application/octet-stream"},
            )
        except TimeoutException:
            raise DekerTimeoutServer(
                message=f"Timeout on {self.type.name} read {array}",
            )

        if response.status_code == STATUS_OK:
            numpy_array = np.frombuffer(response.read(), dtype=array.dtype)
            shape = array[bounds].shape
            if not shape and numpy_array:
                return numpy_array[0]

            return numpy_array.reshape(shape)
        if response.status_code == TIMEOUT:
            raise DekerTimeoutServer(
                message=f"Timeout on {self.type.name} read {array}",
            )
        raise DekerServerError(response, "Couldn't read the array")

    def update(self, array: "BaseArray", bounds: Slice, data: Numeric) -> None:
        """Update array/varray on server.

        :param array: Array/Varray that will be updated
        :param bounds: Part of the array to update
        :param data: Data that will replace current values
        :return:
        """
        bounds = slice_converter[bounds]
        try:
            if hasattr(data, "tolist"):
                data = data.tolist()

            response = self.client.put(
                f"/v1/collection/{array.collection}/{self.type.name}/by-id/{array.id}/subset/{bounds}/data",
                json=data,
            )

        except TimeoutException:
            raise DekerTimeoutServer(
                message=f"Timeout on {self.type.name} update {array}",
            )
        if response.status_code == STATUS_OK:
            return
        if response.status_code == TIMEOUT:
            raise DekerTimeoutServer(
                message=f"Timeout on {self.type.name} update {array}",
            )
        raise DekerServerError(response, "Couldn't update array")

    def clear(self, array: "BaseArray", bounds: Slice) -> None:
        """Clear array/varray.

        :param array: Instance of array/varray
        :param bounds: Part of the array/varray that will be cleared
        """
        return self.update(array, bounds, [])

    def is_deleted(self, array: "BaseArray") -> bool:
        """Check if array/varray was deleted on server.

        :param array: Instance of array/varray
        """
        return False

    def get_by_primary_attributes(
        self,
        primary_attributes: dict,
        schema: "BaseArraysSchema",
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"],
    ) -> Optional["BaseArray"]:
        """Get array by given primary attribute.

        :param primary_attributes: Key attributes
        :param schema: Array or Varray schema
        :param collection: Collection instance
        :param array_adapter: Adapter of arrays
        :param varray_adapter: Adapter of varrays
        """
        attrs = []
        primary_attributes = convert_datetime_attrs_to_iso(primary_attributes)
        for attr in schema.attributes:
            if not attr.primary:
                continue

            if attr.dtype == tuple:
                attrs.append(str(tuple(primary_attributes[attr.name])))  # type: ignore[index]
            else:
                attrs.append(str(primary_attributes[attr.name]))  # type: ignore[index]

        primary_path = "/".join(attrs)
        response = self.client.get(
            f"{self.collection_path.raw_url}/{self.type.name}/by-primary/{primary_path}",
        )
        if response.status_code == NOT_FOUND:
            return None
        if response.status_code == STATUS_OK:
            return create_array_from_meta(
                type=self.type,
                collection=collection,
                meta=response.json(),
                array_adapter=array_adapter,
                varray_adapter=varray_adapter,
            )
        return None

    def get_by_id(
        self,
        id_: str,
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"],
    ) -> Optional["BaseArray"]:
        """Get and array/varray by given ID.

        :param id_: ID of array/varray
        :param collection: Deker's collection instance
        :param array_adapter: Array adapter
        :param varray_adapter: Varray adapter
        :return:
        """
        response = self.client.get(
            f"{self.collection_path.raw_url}/{self.type.name}/by-id/{id_}",
        )
        if response.status_code == NOT_FOUND:
            return None
        if response.status_code == STATUS_OK:
            return create_array_from_meta(
                type=self.type,
                collection=collection,
                meta=response.json(),
                array_adapter=array_adapter,
                varray_adapter=varray_adapter,
            )

        raise DekerServerError(response, "Couldn't fetch an array")

    def __iter__(self) -> Generator["ArrayMeta", None, None]:
        response = self.client.get(f"{self.collection_path.raw_url}/{self.type.name}s")
        if response.status_code != STATUS_OK:
            raise DekerServerError(response, "Couldn't get list of arrays")

        yield from response.json()



from typing import Any, Type

from deker.errors import DekerMemoryError
from httpx import Client, Response

from deker_server_adapters.consts import (
    CONTENT_TOO_LARGE,
    EXCEPTION_CLASS_PARAM_NAME,
    RATE_ERROR_MESSAGE,
    TOO_LARGE_ERROR_MESSAGE,
    TOO_MANY_REQUESTS,
)
from deker_server_adapters.errors import DekerBaseRateLimitError, DekerDataPointsLimitError, DekerRateLimitError


def rate_limit_err(response: Response, message: str, class_: Type[DekerBaseRateLimitError]) -> None:
    """Raise an error with rate limit parameters.

    This function incapsulates logic of fetching rates from headers

    :param response: HttpxResponse object
    :param message: Message of the exception
    :param class_: Class (Type) of the exception
    """
    limit = response.headers.get("RateLimit-Limit")
    remaining = response.headers.get("RateLimit-Remaining")
    reset = response.headers.get("RateLimit-Reset")

    raise class_(
        message=message,
        limit=int(limit) if limit else None,
        remaining=int(remaining) if remaining else None,
        reset=int(reset) if reset else None,
    )


class HttpxClient(Client):
    """Wrapper around HttpxClient."""

    def request(self, *args: Any, **kwargs: Any) -> Response:
        """Override httpx method to handle rate errors.

        :param args: arguments to request
        :param kwargs: keyword arguments to request
        """
        response = super().request(*args, **kwargs)
        if response.status_code == TOO_MANY_REQUESTS:
            rate_limit_err(response=response, message=RATE_ERROR_MESSAGE, class_=DekerRateLimitError)
        elif (
            response.status_code == CONTENT_TOO_LARGE
            and response.json().get(EXCEPTION_CLASS_PARAM_NAME) != DekerMemoryError.__name__
        ):
            rate_limit_err(response=response, message=TOO_LARGE_ERROR_MESSAGE, class_=DekerDataPointsLimitError)

        return response



from json import JSONDecodeError
from typing import Optional

from deker.errors import DekerBaseApplicationError
from httpx import Response


class DekerServerError(DekerBaseApplicationError):
    """Any non-convertable-to-exception status (like 500)."""

    MAX_ERROR_TEXT_SIZE = 100

    def _make_response_from_response(
        self,
        response: Response,
        message: Optional[str] = None,
    ) -> str:
        """Make an Exception message based on response and provided text.

        :param response: Response from deker server
        :param message: Provided message
        """
        try:
            server_message = response.json()
        except JSONDecodeError:
            suffix = "..." if len(response.content) > self.MAX_ERROR_TEXT_SIZE else ""
            server_message = f"{response.content[:self.MAX_ERROR_TEXT_SIZE]!r}{suffix}"

        message = message or ""
        return f"{message} \nResponse: status={response.status_code}, message={server_message}"

    def __init__(
        self,
        response: Optional[Response] = None,
        message: Optional[str] = None,
    ) -> None:
        if response:
            self.message = self._make_response_from_response(response, message)
        else:
            self.message = message if message else ""
        super().__init__(self.message)


class DekerTimeoutServer(DekerServerError):
    """If we hit timeout on server."""


class DekerBaseRateLimitError(DekerBaseApplicationError):
    """If we exceed request limit or number of data points is too large."""

    limit: Optional[int]
    remaining: Optional[int]
    reset: Optional[int]

    def __init__(
        self, message: str, limit: Optional[int] = None, remaining: Optional[int] = None, reset: Optional[int] = None
    ):
        self.limit = limit
        self.remaining = remaining
        self.reset = reset
        self.message = message
        super().__init__(message)


class DekerRateLimitError(DekerBaseRateLimitError):
    """If we exceed request limit."""


class DekerDataPointsLimitError(DekerBaseRateLimitError):
    """If number of data points is too large."""



from collections.abc import Generator

from deker import Collection
from deker.ABC import BaseCollectionAdapter
from deker.errors import DekerCollectionAlreadyExistsError, DekerCollectionNotExistsError
from deker.uri import Uri
from httpx import Client

from deker_server_adapters.base import ServerAdapterMixin
from deker_server_adapters.consts import BAD_REQUEST, COLLECTION_NAME_PARAM, NOT_FOUND, STATUS_CREATED, STATUS_OK
from deker_server_adapters.errors import DekerServerError


class ServerCollectionAdapter(ServerAdapterMixin, BaseCollectionAdapter):
    """Server realization for collection adapter."""

    COLLECTION_URL_PREFIX = "v1/collection"
    COLLECTIONS_URL_PREFIX = "v1/collections"
    client: Client
    metadata_version = "0.2"

    def delete(self, collection: "Collection") -> None:
        """Delete collection from server.

        :param collection: Deker's collection instance
        """
        response = self.client.delete(
            f"/{self.COLLECTION_URL_PREFIX}/{collection.name}",
        )

        if response.status_code == STATUS_OK:
            return

        if response.status_code == NOT_FOUND:
            raise DekerCollectionNotExistsError

        raise DekerServerError(response, "Collection wasn't deleted")

    def create(self, collection: "Collection") -> None:
        """Create collection.

        :param collection: Deker's collection instance
        """
        data = collection.as_dict
        response = self.client.post(f"/{self.COLLECTIONS_URL_PREFIX}", json=data)

        if response.status_code == STATUS_CREATED:
            return

        if response.status_code == BAD_REQUEST and COLLECTION_NAME_PARAM in response.json().get("parameters", []):
            raise DekerCollectionAlreadyExistsError

        raise DekerServerError(response, "Collection wasn't created")

    def read(self, name: str) -> dict:
        """Read collection metadata.

        :param name: Name of collection
        """
        response = self.client.get(f"/{self.COLLECTION_URL_PREFIX}/{name}")

        if response.status_code == STATUS_OK:
            return response.json()

        if response.status_code == NOT_FOUND:
            raise DekerCollectionNotExistsError

        raise DekerServerError(response, f"Couldn't get the collection {name=}")

    def clear(self, collection: "Collection") -> None:
        """Clear collection.

        :param collection: Instance of Deker's collection
        :return:
        """
        array_type = "varrays" if collection.varray_schema else "arrays"
        response = self.client.put(
            f"/{self.COLLECTION_URL_PREFIX}/{collection.name}/{array_type}",
            json=[],
        )

        if response.status_code == NOT_FOUND:
            raise DekerCollectionNotExistsError

        if response.status_code == STATUS_OK:
            return

        raise DekerServerError(
            response,
            f"Couldn't clear the collection {collection.name}",
        )

    @property
    def collections_resource(self) -> Uri:
        """Full path with to collections resource."""
        return self.uri / self.COLLECTION_URL_PREFIX

    def is_deleted(self, collection: Collection) -> bool:
        """Check if method is deleted.

        :param collection: Collection instance
        :return:
        """
        return False

    def __iter__(self) -> Generator[dict, None, None]:
        all_collections_response = self.client.get(f"/{self.COLLECTIONS_URL_PREFIX}")
        if all_collections_response.status_code != STATUS_OK:
            raise DekerServerError(
                all_collections_response,
                "Couldn't get list of collections",
            )

        results = all_collections_response.json()
        yield from results



from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING, Type, Union

from deker.ABC.base_adapters import BaseArrayAdapter, BaseStorageAdapter
from deker.ctx import CTX
from httpx import Client

from deker_server_adapters.base import ServerArrayAdapterMixin
from deker_server_adapters.consts import STATUS_OK, ArrayType
from deker_server_adapters.errors import DekerServerError


if TYPE_CHECKING:
    from deker import Collection
    from deker.uri import Uri


class ServerArrayAdapter(ServerArrayAdapterMixin, BaseArrayAdapter):
    """Server realization for array adapter."""

    client: Client
    type = ArrayType.array
    executor: ThreadPoolExecutor

    def __init__(
        self,
        collection_path: Union["Uri", Path],
        ctx: CTX,
        executor: ThreadPoolExecutor,
        storage_adapter: Type["BaseStorageAdapter"],
    ) -> None:
        super().__init__(
            collection_path,
            ctx,
            executor,
            storage_adapter,
        )
        self.executor = executor

    def delete_all_by_vid(self, vid: str, collection: "Collection") -> None:
        """Delete all arrays that have that vid.

        :param vid: Id of virtual array
        :param collection: Instance of deker's collection
        """
        all_arrays_response = self.client.get(f"{self.collection_path.raw_url}/arrays/")
        if all_arrays_response.status_code == STATUS_OK:
            arrays_with_vid = [
                array["id"] for array in all_arrays_response.json() if array["primary_attributes"].get("vid") == vid
            ]

            def delete_by_id(id_: str) -> bool:
                """Try to delete array with given id.

                :param id_: Id of array.
                """
                try:
                    response = self.client.delete(f"{self.collection_path.raw_url}/array/by-id/{id_}")
                    if response.status_code != STATUS_OK:
                        raise DekerServerError(response, "Couldn't delete the array")
                except Exception:
                    return False
                return True

            results = self.executor.map(delete_by_id, arrays_with_vid)
            errors = []
            for i, result in enumerate(results):
                if not result:
                    errors.append(arrays_with_vid[i])
            if errors:
                raise DekerServerError(
                    None,
                    f"Not all arrays were deleted. Failed arrays: {errors}",
                )
            return
        raise DekerServerError(all_arrays_response, "Couldn't fetch list of the arrays")



from enum import Enum

from deker import Array, VArray


STATUS_OK = 200
STATUS_CREATED = 201
BAD_REQUEST = 400
NOT_FOUND = 404
TIMEOUT = 504
CONTENT_TOO_LARGE = 413
TOO_MANY_REQUESTS = 429

# Name of the parameter in the server response
COLLECTION_NAME_PARAM = "collection_name"

RATE_ERROR_MESSAGE = "Too many requests, try again later"
TOO_LARGE_ERROR_MESSAGE = "Requested object is too large, use smaller subset"

EXCEPTION_CLASS_PARAM_NAME = "class"


class ArrayType(Enum):
    """Enum for array/varray values."""

    array = Array
    varray = VArray
